﻿
using System;
using System.Drawing.Imaging;
using System.Drawing;
using SoftoriaPole.Math;
namespace SoftoriaPole.Geometry.Utils
{
    // ------------------------------------------------------------

    // This implementation correctly handles duplicate points, and
    // multiple points with the same x-coordinate.

    // 1. Sort the points lexicographically by increasing (x,y), thus 
    //    finding also a leftmost point L and a rightmost point R.
    // 2. Partition the point set into two lists, upper and lower, according as 
    //    point is above or below the segment LR.  The upper list begins with 
    //    L and ends with R; the lower list begins with R and ends with L.
    // 3. Traverse the point lists clockwise, eliminating all but the extreme
    //    points (thus eliminating also duplicate points).
    // 4. Eliminate L from lower and R from upper, if necessary.
    // 5. Join the point lists (in clockwise order) in an array.
    [Serializable]
    public class Convexhull
    {
        public static VectorD[] convexhull(VectorD[] pts)
        {
            // Sort points lexicographically by increasing (x, y)
            int N = pts.Length;
            if (N <= 1)
                return pts;
            Polysort.Quicksort(pts);
            VectorD left = pts[0];
            VectorD right = pts[N - 1];
            // Partition into lower hull and upper hull
            CDLL lower = new CDLL(left), upper = new CDLL(left);
            for (int i = 0; i < N; i++)
            {
                double det = VectorD.Area2(left, right, pts[i]);
                if (det > 0)
                    upper = upper.Append(new CDLL(pts[i]));
                else if (det < 0)
                    lower = lower.Prepend(new CDLL(pts[i]));
            }
            lower = lower.Prepend(new CDLL(right));
            upper = upper.Append(new CDLL(right)).Next;
            // Eliminate points not on the hull
            eliminate(lower);
            eliminate(upper);
            // Eliminate duplicate endpoints
            if (lower.Prev.val.Equals(upper.val))
                lower.Prev.Delete();
            if (upper.Prev.val.Equals(lower.val))
                upper.Prev.Delete();
            // Join the lower and upper hull
            VectorD[] res = new VectorD[lower.Size() + upper.Size()];
            lower.CopyInto(res, 0);
            upper.CopyInto(res, lower.Size());
            return res;

        }

        // Graham's scan
        private static void eliminate(CDLL start)
        {
            CDLL v = start, w = start.Prev;
            bool fwd = false;
            while (v.Next != start || !fwd)
            {
                if (v.Next == w)
                    fwd = true;
                if (VectorD.Area2(v.val, v.Next.val, v.Next.Next.val) < 0) // right turn
                    v = v.Next;
                else
                {                                       // left turn or straight
                    v.Next.Delete();
                    v = v.Prev;
                }
            }
        }
    }

   
    // ------------------------------------------------------------

    // Circular doubly linked lists of Point
    [Serializable]
    class CDLL
    {
        private CDLL prev, next;     // not null, except in deleted elements
        public VectorD val;

        // A new CDLL node is a one-element circular list
        public CDLL(VectorD val)
        {
            this.val = val; next = prev = this;
        }

        public CDLL Prev
        {
            get { return prev; }
        }

        public CDLL Next
        {
            get { return next; }
        }

        // Delete: adjust the remaining elements, make this one point nowhere
        public void Delete()
        {
            next.prev = prev; prev.next = next;
            next = prev = null;
        }

        public CDLL Prepend(CDLL elt)
        {
            elt.next = this; elt.prev = prev; prev.next = elt; prev = elt;
            return elt;
        }

        public CDLL Append(CDLL elt)
        {
            elt.prev = this; elt.next = next; next.prev = elt; next = elt;
            return elt;
        }

        public int Size()
        {
            int count = 0;
            CDLL node = this;
            do
            {
                count++;
                node = node.next;
            } while (node != this);
            return count;
        }

        public void PrintFwd()
        {
            CDLL node = this;
            do
            {
                Console.WriteLine(node.val);
                node = node.next;
            } while (node != this);
            Console.WriteLine();
        }

        public void CopyInto(VectorD[] vals, int i)
        {
            CDLL node = this;
            do
            {
                vals[i++] = node.val;	// still, implicit checkcasts at runtime 
                node = node.next;
            } while (node != this);
        }
    }

    // ------------------------------------------------------------
    [Serializable]
    class Polysort
    {
        private static void swap(VectorD[] arr, int s, int t)
        {
            VectorD tmp = arr[s]; arr[s] = arr[t]; arr[t] = tmp;
        }
        private static void swap(Ordered[] arr, int s, int t)
        {
            Ordered tmp = arr[s]; arr[s] = arr[t]; arr[t] = tmp;
        }


        // Typed OO-style quicksort a la Hoare/Wirth

        private static void qsort(Ordered[] arr, int a, int b)
        {
            // sort arr[a..b]
            if (a < b)
            {
                int i = a, j = b;
                VectorD x = (VectorD)(arr[(i + j) / 2]);
                do
                {
                    while (arr[i].Less(x)) i++;
                    while (x.Less(arr[j])) j--;
                    if (i <= j)
                    {
                        swap(arr, i, j);
                        i++; j--;
                    }
                } while (i <= j);
                qsort(arr, a, j);
                qsort(arr, i, b);
            }
        }

        public static void Quicksort(Ordered[] arr)
        {
            qsort(arr, 0, arr.Length - 1);
        }
    }
    

    // ------------------------------------------------------------

    /*class TestCH {
      static void Main(string[] args) {
        if (args.Length == 1) {
          int N = int.Parse(args[0]);
          Point[] pts = new Point[N];
          for (int i=0; i<N; i++)
            pts[i] = Point.Random(500, 500);
          Point[] chpts = Convexhull.convexhull(pts);
          Console.WriteLine("Area is " + area(chpts));
          print(chpts);
        } else
          Console.WriteLine("Usage: TestCH <pointcount>\n");
      }

      // The centroid of a point set
      public static Point centroid(Point[] pts) { 
        int N = pts.Length;
        double sumx = 0, sumy = 0;
        for (int i=0; i<N; i++) {
          sumx += pts[i].x;
          sumy += pts[i].y;
        }
        return new Point(sumx/N, sumy/N);
      }

      // The area of a polygon (represented by an array of ordered vertices)
      public static double area(Point[] pts) { 
        int N = pts.Length;
        Point centr = centroid(pts);
        double area2 = 0;
        for (int i=0; i<N; i++) 
          area2 += Point.Area2(centr, pts[i], pts[(i+1)%N]);
        return Math.Abs(area2/2);
      }

      public static void print(Point[] pts) {
        int N = pts.Length;
        for (int i=0; i<N; i++) 
          Console.WriteLine(pts[i]);
      }
    }*/

}